use std::{
    fs::File,
    io::{Read as _, Seek as _, SeekFrom},
};

pub struct Netdev {
    rx_bytes: File,
    tx_bytes: File,
}

#[derive(Debug)]
pub enum NetdevError {
    NotFound { name: String },
    IoError { name: String, err: std::io::Error },
}

impl NetdevError {
    pub fn available_netdevs_msg() -> String {
        let list = Netdev::list()
            .unwrap_or_default()
            .into_iter()
            .map(|dev| format!("    {dev}"))
            .collect::<Vec<_>>()
            .join("\n");

        format!("Available network devices:\n{list}")
    }
}

impl std::fmt::Display for NetdevError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            NetdevError::NotFound { name } => write!(
                f,
                "Netdev '{name}' not found.\n\n{}",
                Self::available_netdevs_msg()
            ),
            NetdevError::IoError { name, err } => {
                write!(f, "I/O Error occurred for netdev '{name}': {err}")
            }
        }
    }
}

impl Netdev {
    const SYS_CLASS_NET: &'static str = "/sys/class/net";

    pub fn from_name(name: &str) -> eyre::Result<Self> {
        let rx_bytes = Self::open_netdev_stat(name, "rx_bytes")?;
        let tx_bytes = Self::open_netdev_stat(name, "tx_bytes")?;

        Ok(Self { rx_bytes, tx_bytes })
    }

    fn open_netdev_stat(name: &str, stat: &str) -> eyre::Result<File> {
        let sys_class_net = Self::SYS_CLASS_NET;
        let path = format!("{sys_class_net}/{name}/statistics/{stat}");
        match File::open(&path) {
            Ok(file) => Ok(file),
            Err(e) => {
                if e.kind() == std::io::ErrorKind::NotFound {
                    Err(eyre::eyre!(NetdevError::NotFound {
                        name: name.to_string(),
                    }))
                } else {
                    Err(eyre::eyre!(NetdevError::IoError {
                        name: name.to_string(),
                        err: e,
                    }))
                }
            }
        }
    }

    fn read_bytes(file: &mut File) -> eyre::Result<u64> {
        file.seek(SeekFrom::Start(0))?;
        let mut buf = String::new();
        file.read_to_string(&mut buf)?;
        Ok(buf.trim().parse().unwrap_or(0))
    }

    pub fn rx_bytes(&mut self) -> eyre::Result<u64> {
        Self::read_bytes(&mut self.rx_bytes)
    }

    pub fn tx_bytes(&mut self) -> eyre::Result<u64> {
        Self::read_bytes(&mut self.tx_bytes)
    }

    pub fn list() -> eyre::Result<Vec<String>> {
        let mut entries = std::fs::read_dir(Self::SYS_CLASS_NET)?
            .filter_map(|entry| {
                entry
                    .ok()
                    .and_then(|entry| entry.file_name().into_string().ok())
            })
            .collect::<Vec<_>>();

        entries.sort();
        Ok(entries)
    }
}
